/*
 * ExpansionLayout.java
 *
 * Created on September 4, 2001, 10:43 AM
 *
 * See LICENSE file for license conditions.
 */

package residue;

import java.awt.*;

/**
 * A simple layout manager to improve the look of the GUI.
 *
 * All components in the Layout are put in one row.
 *
 * Each component is given its preferred height, and
 * centered vertically, possibly cropping elements.
 *
 * Width is determined by expanding or contracting proportionally
 * between min, preferred, and max values.  If the size given is
 * less than the minimum size, or greater than the maximum size,
 * it aligns the components either to the left, right, or center,
 * as specified in the constructor in the setAlignment method.
 *
 * @author  chowells
 * @version 
 */
public class ExpansionLayout implements LayoutManager
{
	private int hgap;
	private int vgap;
	
	private int align;
	
	/**
	 * The constant used to represent centered alignment.
	 */
	public static final int CENTERED = 0;
	
	/**
	 * The constant used to represent left alignment.
	 */
	public static final int LEFT = 1;
	
	/**
	 * The constant used to represent right alignment.
	 */
	public static final int RIGHT = 2;
	
	/**
	 * Creates a new ExpansionLayout with the given horizontal and
	 * vertical gaps, and the specified alignment.
	 */
	public ExpansionLayout(int horizontalgap, int verticalgap, int alignment)
	{
		setHorizontalGap(horizontalgap);
		setVerticalGap(verticalgap);
		setAlignment(alignment);
	} // ExpansionLayout
	
	/**
	 * Creates a new ExpansionLayout with the given horizontal and
	 * vertical gaps, with CENTERED alignment.
	 */
	public ExpansionLayout(int horizontalgap, int verticalgap)
	{
		this(horizontalgap, verticalgap, CENTERED);
	} // ExpansionLayout
	
	/**
	 * Creates a new ExpansionLayout with the given alignment, and
	 * with horizontal and vertical gap values of 5.
	 */
	public ExpansionLayout(int alignment)
	{
		this(5, 5, alignment);
	} // ExpansionLayout
	
	/**
	 * Creates a new ExpansionLayout with centered alignment, and
	 * horizontal and vertical gap values of 5.
	 */
	public ExpansionLayout()
	{
		this(5, 5, CENTERED);
	} // ExpansionLayout
	
	/**
	 * Sets the alignment of this ExpansionLayout.
	 */
	public void setAlignment(int alignment)
	{
		if (alignment < CENTERED || alignment > RIGHT)
			throw new IllegalArgumentException("Invalid alignment constant.");
		
		align = alignment;
	} // setAlignment
	
	/**
	 * Returns the alignment setting of this ExpansionLayout.
	 */
	public int getAlignment()
	{
		return align;
	} // getAlignment
	
	/**
	 * Sets the horizontal gap of this ExpansionLayout.
	 */
	public void setHorizontalGap(int horizontalgap)
	{
		hgap = horizontalgap;
	} // setHorizontalGap
	
	/**
	 * Returns the horizontal gap of this ExpansionLayout.
	 */
	public int getHorizontalGap()
	{
		return hgap;
	} // getHorizontalGap
	
	/**
	 * Sets the vertical gap of this ExpansionLayout.
	 */
	public void setVerticalGap(int verticalgap)
	{
		vgap = verticalgap;
	} // setVerticalGap
	
	/**
	 * Returns the vertical gap of this ExpansionLayout.
	 */
	public int getVerticalGap()
	{
		return vgap;
	} // getVerticalGap
	
	/**
	 * Lays out the container, as described before.
	 */
	public void layoutContainer(Container target)
	{
		synchronized (target.getTreeLock())
		{
			Insets insets = target.getInsets();
			int maxheight = target.getHeight() - (insets.top + insets.bottom + vgap * 2);
			
			int visible = 0;
			
			for (int i = 0; i < target.getComponentCount(); i++)
				if (target.getComponent(i).isVisible()) visible++;
			
			int count = 0;
			Component [] components = new Component[visible];
			for (int i = 0; i < target.getComponentCount(); i++)
				if (target.getComponent(i).isVisible())
					components[count++] = target.getComponent(i);
			
			int available = target.getWidth() - (insets.left + insets.right + (1 + visible) * hgap);
			
			int min = 0;
			int pref = 0;
			long max = 0;
			
			int [][] sizes = new int[visible][3];
			for (int i = 0; i < visible; i++)
			{
				sizes[i][0] = components[i].getMinimumSize().width;
				min += sizes[i][0];
				sizes[i][1] = Math.max(sizes[i][0], components[i].getPreferredSize().width);
				pref += sizes[i][1];
				sizes[i][2] = Math.max(sizes[i][1], components[i].getMaximumSize().width);
				max += sizes[i][2];
			}
			
			if (max > Integer.MAX_VALUE) max = Integer.MAX_VALUE;
			
			double ratio;
			int offset;
			int lower;
			
			if (available < min)
			{
				ratio = 0.0;
				lower = 0;
				switch (align)
				{
					case CENTERED:
						offset = insets.left + hgap +
							(target.getSize().width - minimumLayoutSize(target).width) / 2;
						break;
					case LEFT:
						offset = insets.left + hgap;
						break;
					case RIGHT:
						offset = insets.left + hgap + 
							target.getSize().width - minimumLayoutSize(target).width;
						break;
					default:
						throw new IllegalStateException("Invalid alignment");
				} // switch
			}
			else if (available < pref)
			{
				ratio = (available - min) / ((double)(pref - min));
				lower = 0;
				offset = insets.left + hgap;
			}
			else if (available < max)
			{
				ratio = (available - pref) / ((double)(max - pref));
				lower = 1;
				offset = insets.left + hgap;
			}
			else
			{
				ratio = 1.0;
				lower = 1;
				switch (align)
				{
					case CENTERED:
						offset = insets.left + hgap +
							(target.getSize().width - maximumLayoutSize(target).width) / 2;
						break;
					case LEFT:
						offset = insets.left + hgap;
						break;
					case RIGHT:
						offset = insets.left + hgap +
							target.getSize().width - maximumLayoutSize(target).width;
						break;
					default:
						throw new IllegalStateException("Invalid alignment");
				} // switch
			}
			
			for (int i = 0; i < visible; i++)
			{
				int width = sizes[i][lower];
				width += (int)(ratio * (sizes[i][lower + 1] - sizes[i][lower]));
				
				int height = components[i].getPreferredSize().height;
				
				components[i].setSize(width, height);
				components[i].setLocation(offset, insets.top + vgap + (maxheight - height) / 2);
				
				offset += hgap + width;
			} // for
			
		} // synchronized
	} // layoutContainer
	
	/**
	 * Returns this layout's maximum size for the given container,
	 * without padding components.
	 */
	public Dimension maximumLayoutSize(Container target)
	{
		long height = 0;
		long width = 0;
		
		synchronized (target.getTreeLock())
		{
			int nmembers = target.getComponentCount();
			
			for (int i = 0 ; i < nmembers ; i++)
			{
				Component m = target.getComponent(i);
				if (m.isVisible())
				{
					Dimension dm = m.getMaximumSize();
					Dimension dp = m.getPreferredSize();
					height = Math.max(height, dp.height);
					
					width += dm.width + hgap;
				}
			}
			
			Insets insets = target.getInsets();
			width += insets.left + insets.right + hgap;
			height += insets.top + insets.bottom + vgap * 2;
		} // synchronized
		
		if (height > Integer.MAX_VALUE) height = Integer.MAX_VALUE;
		if (width > Integer.MAX_VALUE) width = Integer.MAX_VALUE;
		
		return new Dimension((int)width, (int)height);
	} // maximumLayoutSize
	
	/**
	 * Returns this layout's preferred size for the given Container.
	 */
	public Dimension preferredLayoutSize(Container target)
	{
		long height = 0;
		long width = 0;
		
		synchronized (target.getTreeLock())
		{
			int nmembers = target.getComponentCount();
			
			for (int i = 0 ; i < nmembers ; i++)
			{
				Component m = target.getComponent(i);
				if (m.isVisible())
				{
					Dimension d = m.getPreferredSize();
					height = Math.max(height, d.height);
					
					width += d.width + hgap;
				}
			}
			
			Insets insets = target.getInsets();
			width += insets.left + insets.right + hgap;
			height += insets.top + insets.bottom + vgap * 2;
		} // synchronized
		
		if (height > Integer.MAX_VALUE) height = Integer.MAX_VALUE;
		if (width > Integer.MAX_VALUE) width = Integer.MAX_VALUE;
		
		return new Dimension((int)width, (int)height);
	} // preferredLayoutSize
	
	/**
	 * Returns this layout's minimum size for the given container.
	 */
	public Dimension minimumLayoutSize(Container target)
	{
		long width = 0;
		long height = 0;
		
		synchronized (target.getTreeLock())
		{
			int nmembers = target.getComponentCount();
			
			for (int i = 0 ; i < nmembers ; i++)
			{
				Component m = target.getComponent(i);
				if (m.isVisible())
				{
					Dimension dm = m.getMinimumSize();
					Dimension dp = m.getPreferredSize();
					height = Math.max(height, dp.height);
					
					width += dm.width + hgap;
				}
			}
			
			Insets insets = target.getInsets();
			width += insets.left + insets.right + hgap;
			height += insets.top + insets.bottom + vgap * 2;
		} // synchronized
		
		if (height > Integer.MAX_VALUE) height = Integer.MAX_VALUE;
		if (width > Integer.MAX_VALUE) width = Integer.MAX_VALUE;
		
		return new Dimension((int)width, (int)height);
	} // minimumLayoutSize
	
	/**
	 * This method isn't used in this class.
	 */
	public void addLayoutComponent(String str, Component component) {}
	
	/**
	 * This method isn't used in this class.
	 */
	public void removeLayoutComponent(Component component) {}
	
} // class ExpansionLayout
